@@ -17,6 +17,7 @@ use {
     elliptic_curve::{
         group::Curve as _,
         ops::{Invert, LinearCombination, Reduce},
+        subtle::CtOption,
         AffineArithmetic, AffineXCoordinate, Field, FieldBytes, Group, ProjectiveArithmetic,
         ProjectivePoint, Scalar, ScalarArithmetic,
     },
@@ -34,10 +35,18 @@ use crate::{
     Signature,
 };
 
-#[cfg(all(feature = "sign"))]
+#[cfg(all(feature = "arithmetic", feature = "digest"))]
+use signature::digest::FixedOutput;
+
+#[cfg(all(feature = "rfc6979"))]
 use {
-    elliptic_curve::{ff::PrimeField, zeroize::Zeroizing, NonZeroScalar, ScalarCore},
-    signature::digest::{BlockInput, FixedOutput, Reset, Update},
+    elliptic_curve::ScalarCore,
+    signature::digest::{
+        block_buffer::Eager,
+        core_api::{BlockSizeUser, BufferKindUser, CoreProxy, FixedOutputCore},
+        generic_array::typenum::{IsLess, Le, NonZero, U256},
+        HashMarker, OutputSizeUser,
+    },
 };
 
 /// Try to sign the given prehashed message using ECDSA.
@@ -56,13 +65,8 @@ where
     /// Accepts the following arguments:
     ///
     /// - `k`: ephemeral scalar value. MUST BE UNIFORMLY RANDOM!!!
-    /// - `z`: scalar computed from a hashed message digest to be signed.
-    ///   MUST BE OUTPUT OF A CRYPTOGRAPHICALLY SECURE DIGEST ALGORITHM!!!
-    ///
-    /// # Computing the `hashed_msg` scalar
-    ///
-    /// To compute a [`Scalar`] from a message digest, use the [`Reduce`] trait
-    /// on the computed digest, e.g. `Scalar::from_be_bytes_reduced`.
+    /// - `z`: message digest to be signed. MUST BE OUTPUT OF A CRYPTOGRAPHICALLY
+    ///        SECURE DIGEST ALGORITHM!!!
     ///
     /// # Returns
     ///
@@ -72,26 +76,28 @@ where
     fn try_sign_prehashed<K>(
         &self,
         k: K,
-        z: Scalar<C>,
+        z: FieldBytes<C>,
     ) -> Result<(Signature<C>, Option<RecoveryId>)>
     where
-        K: Borrow<Self> + Invert<Output = Self>,
+        K: Borrow<Self> + Invert<Output = CtOption<Self>>,
     {
         if k.borrow().is_zero().into() {
             return Err(Error::new());
         }
 
+        let z = Self::from_be_bytes_reduced(z);
+
         // Compute scalar inversion of 𝑘
         let k_inv = Option::<Scalar<C>>::from(k.invert()).ok_or_else(Error::new)?;
 
-        // Compute 𝐑 = 𝑘×𝑮
+        // Compute 𝑹 = 𝑘×𝑮
         let R = (C::ProjectivePoint::generator() * k.borrow()).to_affine();
 
-        // Lift x-coordinate of 𝐑 (element of base field) into a serialized big
+        // Lift x-coordinate of 𝑹 (element of base field) into a serialized big
         // integer, then reduce it into an element of the scalar field
         let r = Self::from_be_bytes_reduced(R.x());
 
-        // Compute `s` as a signature over `r` and `z`.
+        // Compute 𝒔 as a signature over 𝒓 and 𝒛.
         let s = k_inv * (z + (r * self));
 
         if s.is_zero().into() {
@@ -101,6 +107,69 @@ where
         // TODO(tarcieri): support for computing recovery ID
         Ok((Signature::from_scalars(r, s)?, None))
     }
+
+    /// Try to sign the given message digest deterministically using the method
+    /// described in [RFC6979] for computing ECDSA ephemeral scalar `k`.
+    ///
+    /// Accepts the following parameters:
+    /// - `z`: message digest to be signed.
+    /// - `ad`: optional additional data, e.g. added entropy from an RNG
+    ///
+    /// [RFC6979]: https://datatracker.ietf.org/doc/html/rfc6979
+    #[cfg(all(feature = "rfc6979"))]
+    #[cfg_attr(docsrs, doc(cfg(feature = "rfc6979")))]
+    fn try_sign_prehashed_rfc6979<D>(
+        &self,
+        z: FieldBytes<C>,
+        ad: &[u8],
+    ) -> Result<(Signature<C>, Option<RecoveryId>)>
+    where
+        Self: From<ScalarCore<C>>,
+        C::UInt: for<'a> From<&'a Self>,
+        D: CoreProxy + FixedOutput<OutputSize = FieldSize<C>>,
+        D::Core: BlockSizeUser
+            + BufferKindUser<BufferKind = Eager>
+            + Clone
+            + Default
+            + FixedOutputCore
+            + HashMarker
+            + OutputSizeUser<OutputSize = D::OutputSize>,
+        <D::Core as BlockSizeUser>::BlockSize: IsLess<U256>,
+        Le<<D::Core as BlockSizeUser>::BlockSize, U256>: NonZero,
+    {
+        let x = C::UInt::from(self);
+        let k = rfc6979::generate_k::<D, C::UInt>(&x, &C::ORDER, &z, ad);
+        let k = Self::from(ScalarCore::<C>::new(*k).unwrap());
+        self.try_sign_prehashed(k, z)
+    }
+
+    /// Try to sign the given digest instance using the method described in
+    /// [RFC6979].
+    ///
+    /// [RFC6979]: https://datatracker.ietf.org/doc/html/rfc6979
+    #[cfg(all(feature = "rfc6979"))]
+    #[cfg_attr(docsrs, doc(cfg(feature = "rfc6979")))]
+    fn try_sign_digest_rfc6979<D>(
+        &self,
+        msg_digest: D,
+        ad: &[u8],
+    ) -> Result<(Signature<C>, Option<RecoveryId>)>
+    where
+        Self: From<ScalarCore<C>>,
+        C::UInt: for<'a> From<&'a Self>,
+        D: CoreProxy + FixedOutput<OutputSize = FieldSize<C>>,
+        D::Core: BlockSizeUser
+            + BufferKindUser<BufferKind = Eager>
+            + Clone
+            + Default
+            + FixedOutputCore
+            + HashMarker
+            + OutputSizeUser<OutputSize = D::OutputSize>,
+        <D::Core as BlockSizeUser>::BlockSize: IsLess<U256>,
+        Le<<D::Core as BlockSizeUser>::BlockSize, U256>: NonZero,
+    {
+        self.try_sign_prehashed_rfc6979::<D>(msg_digest.finalize_fixed(), ad)
+    }
 }
 
 /// Verify the given prehashed message using ECDSA.
@@ -120,11 +189,13 @@ where
     ///
     /// Accepts the following arguments:
     ///
-    /// - `z`: prehashed message to be verified
+    /// - `z`: message digest to be verified. MUST BE OUTPUT OF A
+    ///        CRYPTOGRAPHICALLY SECURE DIGEST ALGORITHM!!!
     /// - `sig`: signature to be verified against the key and message
-    fn verify_prehashed(&self, z: Scalar<C>, sig: &Signature<C>) -> Result<()> {
+    fn verify_prehashed(&self, z: FieldBytes<C>, sig: &Signature<C>) -> Result<()> {
+        let z = Scalar::<C>::from_be_bytes_reduced(z);
         let (r, s) = sig.split_scalars();
-        let s_inv = Option::<Scalar<C>>::from(s.invert()).ok_or_else(Error::new)?;
+        let s_inv = *s.invert();
         let u1 = z * s_inv;
         let u2 = *r * s_inv;
         let x = ProjectivePoint::<C>::lincomb(
@@ -142,6 +213,16 @@ where
             Err(Error::new())
         }
     }
+
+    /// Verify message digest against the provided signature.
+    #[cfg(feature = "digest")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "digest")))]
+    fn verify_digest<D>(&self, msg_digest: D, sig: &Signature<C>) -> Result<()>
+    where
+        D: FixedOutput<OutputSize = FieldSize<C>>,
+    {
+        self.verify_prehashed(msg_digest.finalize_fixed(), sig)
+    }
 }
 
 /// Bind a preferred [`Digest`] algorithm to an elliptic curve type.
@@ -170,27 +251,3 @@ where
 {
     type Digest = C::Digest;
 }
-
-/// Deterministically compute ECDSA ephemeral scalar `k` using the method
-/// described in RFC6979.
-///
-/// Accepts the following parameters:
-/// - `x`: secret key
-/// - `z`: message scalar (i.e. message digest reduced mod p)
-/// - `ad`: optional additional data, e.g. added entropy from an RNG
-#[cfg(all(feature = "sign"))]
-#[cfg_attr(docsrs, doc(cfg(feature = "sign")))]
-pub fn rfc6979_generate_k<C, D>(
-    x: &NonZeroScalar<C>,
-    z: &Scalar<C>,
-    ad: &[u8],
-) -> Zeroizing<NonZeroScalar<C>>
-where
-    C: PrimeCurve + ProjectiveArithmetic,
-    D: FixedOutput<OutputSize = FieldSize<C>> + BlockInput + Clone + Default + Reset + Update,
-{
-    // TODO(tarcieri): avoid this conversion
-    let x = Zeroizing::new(ScalarCore::<C>::from(x));
-    let k = rfc6979::generate_k::<D, C::UInt>(x.as_uint(), &C::ORDER, &z.to_repr(), ad);
-    Zeroizing::new(NonZeroScalar::<C>::from_uint(*k).unwrap())
-}
