@@ -1,4 +1,12 @@
+#![no_std]
+#![cfg_attr(docsrs, feature(doc_cfg))]
 #![doc = include_str!("../README.md")]
+#![doc(
+    html_logo_url = "https://raw.githubusercontent.com/RustCrypto/media/8f1a9894/logo.svg",
+    html_favicon_url = "https://raw.githubusercontent.com/RustCrypto/media/8f1a9894/logo.svg"
+)]
+#![forbid(unsafe_code, clippy::unwrap_used)]
+#![warn(missing_docs, rust_2018_idioms)]
 
 //! ## `serde` support
 //!
@@ -31,16 +39,6 @@
 //! [`ring-compat`]: https://docs.rs/ring-compat
 //! [*ring*]: https://docs.rs/ring
 
-#![no_std]
-#![cfg_attr(docsrs, feature(doc_cfg))]
-#![forbid(unsafe_code, clippy::unwrap_used)]
-#![warn(missing_docs, rust_2018_idioms)]
-#![doc(
-    html_logo_url = "https://raw.githubusercontent.com/RustCrypto/media/8f1a9894/logo.svg",
-    html_favicon_url = "https://raw.githubusercontent.com/RustCrypto/media/8f1a9894/logo.svg",
-    html_root_url = "https://docs.rs/ecdsa/0.13.4"
-)]
-
 #[cfg(feature = "alloc")]
 extern crate alloc;
 
@@ -100,11 +98,11 @@ use {
 };
 
 #[cfg(feature = "serde")]
-use elliptic_curve::serde::{ser, Serialize};
+use serdect::serde::{ser, Serialize};
 
 // TODO(tarcieri): support deserialization with the `arithmetic` feature disabled
 #[cfg(all(feature = "arithmetic", feature = "serde"))]
-use elliptic_curve::serde::{de, Deserialize};
+use serdect::serde::{de, Deserialize};
 
 /// Size of a fixed sized signature for the given elliptic curve.
 pub type SignatureSize<C> = <FieldSize<C> as Add>::Output;
