@@ -11,20 +11,18 @@ use elliptic_curve::{
     sec1::{self, EncodedPoint, FromEncodedPoint, ToEncodedPoint},
     AffinePoint, FieldSize, PointCompression, PrimeCurve, ProjectiveArithmetic, PublicKey, Scalar,
 };
-use signature::{digest::Digest, DigestVerifier, Verifier};
+use signature::digest::{Digest, FixedOutput};
+use signature::{DigestVerifier, Verifier};
 
 #[cfg(feature = "pkcs8")]
-use elliptic_curve::{
-    pkcs8::{self, DecodePublicKey},
-    AlgorithmParameters,
-};
+use elliptic_curve::pkcs8::{self, AssociatedOid, DecodePublicKey};
 
 #[cfg(feature = "pem")]
 use core::str::FromStr;
 
 #[cfg(all(feature = "pem", feature = "serde"))]
 #[cfg_attr(docsrs, doc(cfg(all(feature = "pem", feature = "serde"))))]
-use elliptic_curve::serde::{de, ser, Deserialize, Serialize};
+use serdect::serde::{de, ser, Deserialize, Serialize};
 
 /// ECDSA verification key (i.e. public key). Generic over elliptic curves.
 ///
@@ -61,6 +59,16 @@ where
             .map_err(|_| Error::new())
     }
 
+    /// Initialize [`VerifyingKey`] from an affine point.
+    ///
+    /// Returns an [`Error`] if the given affine point is the additive identity
+    /// (a.k.a. point at infinity).
+    pub fn from_affine(affine: AffinePoint<C>) -> Result<Self> {
+        Ok(Self {
+            inner: PublicKey::from_affine(affine).map_err(|_| Error::new())?,
+        })
+    }
+
     /// Initialize [`VerifyingKey`] from an [`EncodedPoint`].
     pub fn from_encoded_point(public_key: &EncodedPoint<C>) -> Result<Self> {
         Option::from(PublicKey::<C>::from_encoded_point(public_key))
@@ -80,27 +88,26 @@ impl<C> Copy for VerifyingKey<C> where C: PrimeCurve + ProjectiveArithmetic {}
 impl<C, D> DigestVerifier<D, Signature<C>> for VerifyingKey<C>
 where
     C: PrimeCurve + ProjectiveArithmetic,
-    D: Digest<OutputSize = FieldSize<C>>,
+    D: Digest + FixedOutput<OutputSize = FieldSize<C>>,
     AffinePoint<C>: VerifyPrimitive<C>,
     Scalar<C>: Reduce<C::UInt>,
     SignatureSize<C>: ArrayLength<u8>,
 {
     fn verify_digest(&self, msg_digest: D, signature: &Signature<C>) -> Result<()> {
-        let scalar = Scalar::<C>::from_be_bytes_reduced(msg_digest.finalize());
-        self.inner.as_affine().verify_prehashed(scalar, signature)
+        self.inner.as_affine().verify_digest(msg_digest, signature)
     }
 }
 
 impl<C> Verifier<Signature<C>> for VerifyingKey<C>
 where
     C: PrimeCurve + ProjectiveArithmetic + DigestPrimitive,
-    C::Digest: Digest<OutputSize = FieldSize<C>>,
+    C::Digest: Digest + FixedOutput<OutputSize = FieldSize<C>>,
     AffinePoint<C>: VerifyPrimitive<C>,
     Scalar<C>: Reduce<C::UInt>,
     SignatureSize<C>: ArrayLength<u8>,
 {
     fn verify(&self, msg: &[u8], signature: &Signature<C>) -> Result<()> {
-        self.verify_digest(C::Digest::new().chain(msg), signature)
+        self.verify_digest(C::Digest::new_with_prefix(msg), signature)
     }
 }
 
@@ -201,7 +208,7 @@ where
 #[cfg_attr(docsrs, doc(cfg(feature = "pkcs8")))]
 impl<C> TryFrom<pkcs8::SubjectPublicKeyInfo<'_>> for VerifyingKey<C>
 where
-    C: PrimeCurve + AlgorithmParameters + ProjectiveArithmetic + PointCompression,
+    C: PrimeCurve + AssociatedOid + ProjectiveArithmetic + PointCompression,
     AffinePoint<C>: FromEncodedPoint<C> + ToEncodedPoint<C>,
     FieldSize<C>: sec1::ModulusSize,
 {
@@ -216,7 +223,7 @@ where
 #[cfg_attr(docsrs, doc(cfg(feature = "pkcs8")))]
 impl<C> DecodePublicKey for VerifyingKey<C>
 where
-    C: PrimeCurve + AlgorithmParameters + ProjectiveArithmetic + PointCompression,
+    C: PrimeCurve + AssociatedOid + ProjectiveArithmetic + PointCompression,
     AffinePoint<C>: FromEncodedPoint<C> + ToEncodedPoint<C>,
     FieldSize<C>: sec1::ModulusSize,
 {
@@ -226,7 +233,7 @@ where
 #[cfg_attr(docsrs, doc(cfg(feature = "pem")))]
 impl<C> FromStr for VerifyingKey<C>
 where
-    C: PrimeCurve + AlgorithmParameters + ProjectiveArithmetic + PointCompression,
+    C: PrimeCurve + AssociatedOid + ProjectiveArithmetic + PointCompression,
     AffinePoint<C>: FromEncodedPoint<C> + ToEncodedPoint<C>,
     FieldSize<C>: sec1::ModulusSize,
 {
@@ -241,7 +248,7 @@ where
 #[cfg_attr(docsrs, doc(cfg(all(feature = "pem", feature = "serde"))))]
 impl<C> Serialize for VerifyingKey<C>
 where
-    C: PrimeCurve + AlgorithmParameters + ProjectiveArithmetic + PointCompression,
+    C: PrimeCurve + AssociatedOid + ProjectiveArithmetic + PointCompression,
     AffinePoint<C>: FromEncodedPoint<C> + ToEncodedPoint<C>,
     FieldSize<C>: sec1::ModulusSize,
 {
@@ -257,7 +264,7 @@ where
 #[cfg_attr(docsrs, doc(cfg(all(feature = "pem", feature = "serde"))))]
 impl<'de, C> Deserialize<'de> for VerifyingKey<C>
 where
-    C: PrimeCurve + AlgorithmParameters + ProjectiveArithmetic + PointCompression,
+    C: PrimeCurve + AssociatedOid + ProjectiveArithmetic + PointCompression,
     AffinePoint<C>: FromEncodedPoint<C> + ToEncodedPoint<C>,
     FieldSize<C>: sec1::ModulusSize,
 {
